from .base import BaseController
from aiohttp import ClientSession
import logging
from typing import List

logger = logging.getLogger("aries_controller.proof")

PRES_PREVIEW = "did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/present-proof/1.0/presentation-preview"

class ProofController(BaseController):

    def __init__(self, admin_url: str, client_session: ClientSession):
        super().__init__(admin_url, client_session)
        self.base_url = "/present-proof"

    def default_handler(self, payload):
        logger.debug("Present Proof Message received", payload)

    async def get_records(self, connection_id: str = None, thread_id: str = None, state: str = None, role: str = None):
        params = {}
        if connection_id:
            params["connection_id"] = connection_id
        if thread_id:
            params["thread_id"] = thread_id
        if state:
            params["state"] = state
        if role:
            params["role"] = role

        return await self.admin_GET(f"{self.base_url}/records", params=params)

    async def get_record_by_id(self, pres_ex_id):
        return await self.admin_GET(f"{self.base_url}/records/{pres_ex_id}")

    # Fetch a single presentation exchange record
    async def get_presentation_credentials(self, pres_ex_id, count: int = None, wql_query: str = None, start: int = None, referent: str = None):
        params = {}
        if count:
            params["count"] = count
        # Not sure what this does
        if wql_query:
            params["extra_query"] = wql_query
        if start:
            params["start"] = start
        # Not sure what this does
        if referent:
            params["referent"] = referent

        return await self.admin_GET(f"{self.base_url}/records/{pres_ex_id}/credentials", params=params)

    # Sends a presentation proposal
    async def send_proposal(self, proposal):

        return await self.admin_POST(f"{self.base_url}/send-proposal", data=proposal)

    # Creates a presentation request not bound to any proposal or existing connection
    async def create_request(self, request):
        # TODO How should proof request object be broken up? Complex.
        #  Do we want user to have to know how to build this object?

        return await self.admin_POST(f"{self.base_url}/create-request", data=request)


    # Sends a free presentation request not bound to any proposal
    async def send_request(self, request):
        return await self.admin_POST(f"{self.base_url}/send-request", data=request)

    async def send_request_for_proposal(self, pres_ex_id, proposal_request):

        return await self.admin_POST(f"{self.base_url}/records/{pres_ex_id}/send-request", data=proposal_request)

    # Send a proof presentation
    async def send_presentation(self, pres_ex_id, presentation):

        return await self.admin_POST(f"{self.base_url}/records/{pres_ex_id}/send-presentation", data=presentation)

    # Verify a received presentation
    async def verify_presentation(self, pres_ex_id):
        return await self.admin_POST(f"{self.base_url}/records/{pres_ex_id}/verify-presentation")

    async def remove_presentation_record(self, pres_ex_id):
        return await self.admin_POST(f"{self.base_url}/records/{pres_ex_id}/remove")

    # def build_proof_request(self, name, version, requested_attributes, requested_predicates):




from .base import BaseController
from aiohttp import ClientSession
import logging
from ..helpers.utils import extract_did, get_schema_details
logger = logging.getLogger("aries_controller.issuer")

CRED_PREVIEW = "did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/issue-credential/1.0/credential-preview"

class IssuerController(BaseController):

    def __init__(self, admin_url: str, client_session: ClientSession, connection_controller,
                 wallet_controller, definition_controller):
        super().__init__(admin_url, client_session)
        self.base_url = "/issue-credential"
        self.connections = connection_controller
        self.wallet = wallet_controller
        self.definitions = definition_controller

    # Fetch all credential exchange records
    async def get_records(self):

        return await self.admin_GET(f"{self.base_url}/records")

    async def get_record_by_id(self, cred_ex_id):
        return await self.admin_GET(f"{self.base_url}/records/{cred_ex_id}")

    # Send holder a credential, automating the entire flow
    # TODO trace=True causes error. Not sure why
    async def send_credential(self, connection_id, schema_id, cred_def_id, attributes, comment: str = "",
                              auto_remove: bool = True, trace: bool = False):

        body = await self.create_credential_body(connection_id, schema_id, cred_def_id, attributes, comment,
                                                 auto_remove, trace)
        return await self.admin_POST(f"{self.base_url}/send", data=body)


    # Send Issuer a credential proposal
    async def send_proposal(self, connection_id, schema_id, cred_def_id, attributes, comment: str = "",
                            auto_remove: bool = True, trace: bool = False):

        body = await self.create_credential_body(connection_id, schema_id, cred_def_id, attributes, comment,
                                                 auto_remove, trace)
        return await self.admin_POST(f"{self.base_url}/send-proposal", data=body)

    async def send_offer(self, connection_id, cred_def_id, attributes, comment: str = "",
                         auto_issue: bool = True, auto_remove: bool = True, trace: bool = False):
        await self.connections.is_active(connection_id)
        offer_body = {
            "cred_def_id": cred_def_id,
            "auto_remove": auto_remove,
            "trace": trace,
            "comment": comment,
            "auto_issue": auto_issue,
            "credential_preview": {
                "@type": CRED_PREVIEW,
                "attributes": attributes
            },
            "connection_id": connection_id
        }
        return await self.admin_POST(f"{self.base_url}/send-offer", data=offer_body)

    # Send holder a credential offer in reference to a proposal with preview
    async def send_offer_for_record(self, cred_ex_id):
        return await self.admin_POST(f"{self.base_url}/records/{cred_ex_id}/send-offer")

    # Send issuer a credential request
    async def send_request_for_record(self, cred_ex_id):
        return await self.admin_POST(f"{self.base_url}/records/{cred_ex_id}/send-request")

    # Send holder a credential
    async def issue_credential(self, cred_ex_id, comment, attributes):
        body = {
            "comment": comment,
            "credential_preview": {
                "@type": CRED_PREVIEW,
                "attributes": attributes
            }
        }
        return await self.admin_POST(f"{self.base_url}/records/{cred_ex_id}/issue", data=body)

    # Store a received credential
    async def store_credential(self, cred_ex_id, credential_id):
        body = {
            "credential_id": credential_id
        }
        return await self.admin_POST(f"{self.base_url}/records/{cred_ex_id}/store", data=body)

    # Revoke and issued credential
    async def revoke_credential(self, rev_reg_id, cred_rev_id, publish: bool = False):
        params = {
            "rev_reg_id": rev_reg_id,
            "cred_reg_id": cred_rev_id,
            "publish": publish
        }
        return await self.admin_POST(f"{self.base_url}/revoke", params=params)

    # Publish pending revocations
    async def publish_revocations(self):
        return await self.admin_POST(f"{self.base_url}/publish-revocations")

    # Remove an existing credential exchange record
    async def remove_record(self, cred_ex_id):
        return await self.admin_POST(f"{self.base_url}/records/{cred_ex_id}/remove")

    # Send a problem report for a credential exchange
    async def problem_report(self, cred_ex_id, explanation: str):
        body = {
            "explain_ltxt": explanation
        }

        return await self.admin_POST(f"{self.base_url}/records/{cred_ex_id}/problem-report", data=body)

    # Used for both send and send-proposal bodies
    async def create_credential_body(self, connection_id, schema_id, cred_def_id, attributes, comment: str = "",
                                     auto_remove: bool = True, trace: bool = False):
        # raises error if connection not active
        await self.connections.is_active(connection_id)

        schema_details = get_schema_details(schema_id)

        issuer_did = extract_did(cred_def_id)

        body = {
            "issuer_did": issuer_did,
            "auto_remove": auto_remove,
            "credential_proposal": {
                "@type": CRED_PREVIEW,
                "attributes": attributes
            },
            "connection_id": connection_id,
            "trace": trace,
            "comment": comment,
            "cred_def_id": cred_def_id,
        }

        credential_body = {**body, **schema_details}
        return credential_body



from .base import BaseController
from aiohttp import ClientSession
import logging

logger = logging.getLogger("aries_controller.definitions")

class DefinitionsController(BaseController):

    def __init__(self, admin_url: str, client_session: ClientSession):
        super().__init__(admin_url, client_session)
        self.base_url = "/credential-definitions"

    async def get_by_id(self, cred_def_id):
        return await self.admin_GET(f"{self.base_url}/{cred_def_id}")

    async def write_cred_def(self, schema_id, tag: str = "default", support_revocation: bool = False):

        body = {
            "schema_id": schema_id,
            "tag": tag,
            "support_revocation": support_revocation
        }

        return await self.admin_POST(f"{self.base_url}", body)

    async def search_created(self, schema_id = None, schema_issuer_did=None, schema_name=None,
                             schema_version=None, issuer_did=None, cred_def_id=None):

        params = {}
        if schema_id:
            params["schema_id"] = schema_id
        if schema_issuer_did:
            params["schema_issuer_did"] = schema_issuer_did
        if schema_version:
            params["schema_version"] = schema_version
        if schema_name:
            params["schema_name"] = schema_name
        if issuer_did:
            params["issuer_did"] = issuer_did
        if cred_def_id:
            params["cred_def_id"] = cred_def_id

        return await self.admin_GET(f"{self.base_url}/created", params=params)




from .base import BaseController
from ..models.connection import Connection

from aiohttp import (
    ClientSession,
)
import logging
import asyncio


logger = logging.getLogger("aries_controller.connections")

class ConnectionsController(BaseController):

    def __init__(self, admin_url: str, client_session: ClientSession):
        super().__init__(admin_url, client_session)
        self.connections = []

    # @staticmethod
    # def default_handler():

    # TODO Add Logging here!
    def default_handler(self, payload):
        connection_id = payload["connection_id"]
        state = payload["state"]
        logger.debug(f"connections hook for {connection_id} with state: {state}")
        for connection in self.connections:
            if connection.id == connection_id:
                connection.update_state(state)
                logger.debug(f"{connection_id} state updated")

    # Combines receive and accept connection api calls
    async def accept_connection(self, invitation):
        response = await self.receive_invitation(invitation)

        accepted = await self.accept_invitation(response["connection_id"])
        return accepted


    ### TODO refactor to extract out generic base - /connections

    async def get_connections(self, alias: str = None, initiator: str = None, invitation_key: str = None, my_did: str = None, state: str = None, their_did: str = None, their_role: str = None):
        params = {}
        if alias:
            params["alias"] = alias
        if initiator:
            params["initiator"] = initiator
        if invitation_key:
            params["invitation_key"] = invitation_key
        if my_did:
            params["my_did"] = my_did
        if state:
            params["state"] = state
        if their_did:
            params["their_did"] = their_did
        if their_role:
            params["their_role"] = their_role

        connections = await self.admin_GET("/connections", params=params)
        return connections

    async def get_connection(self, connection_id: str):
        connection = await self.admin_GET(f"/connections/{connection_id}")
        return connection

    async def create_invitation(self, alias: str = None, auto_accept: bool = None, public: str = None, multi_use: str = None ):
        params = {}
        if alias:
            params["alias"] = alias
        if auto_accept:
            params["auto_accept"] = auto_accept
        if public:
            params["public"] = public
        if multi_use:
            params["multi_use"] = multi_use

        invite_details = await self.admin_POST("/connections/create-invitation", params = params)
        connection = Connection(invite_details["connection_id"], "invitation")
        self.connections.append(connection)
        return invite_details

    async def receive_invitation(self, connection_details: str, alias: str = None, auto_accept: bool = None):
        params = {}
        if alias:
            params["alias"] = alias
        if auto_accept:
            params["auto_accept"] = auto_accept

        response = await self.admin_POST("/connections/receive-invitation", connection_details, params = params)
        connection = Connection(response["connection_id"], response["state"])
        self.connections.append(connection)
        logger.debug("Connection Received - " + connection.id)
        return response

    async def accept_invitation(self, connection_id: str, my_label: str = None, my_endpoint: str = None):
        params = {}
        if my_label:
            params["my_label"] = my_label
        if my_endpoint:
            params["my_endpoint"] = my_endpoint
        response = await self.admin_POST(f"/connections/{connection_id}/accept-invitation",params = params)
        return response


    async def accept_request(self, connection_id: str, my_endpoint: str = None):
        # TODO get if connection_id is in request state, else throw error
        params = {}
        if my_endpoint:
            params["my_endpoint"] = my_endpoint
        connection_ready = await self.check_connection_ready(connection_id, "request")
        if connection_ready:
            response = await self.admin_POST(f"/connections/{connection_id}/accept-request",params = params)
            return response
        else:
            # TODO create proper error classes
            raise Exception("The connection is not in the request state")

    async def establish_inbound(self, connection_id: str, router_conn_id: str):
        response = await self.admin_POST(f"/connections/{connection_id}/establish-inbound/{router_conn_id}")
        return response

    async def remove_connection(self, connection_id):
        response = await self.admin_POST(f"/connections/{connection_id}/remove")
        return response

    async def create_static(self, their_seed, their_label, their_verkey, their_role, my_seed, my_did, their_endpoint, alias, their_did):
        params = {}
        if their_seed:
            params["their_seed"] = their_seed
        if their_label:
            params["their_label"] = their_label
        if their_verkey:
            params["their_verkey"] = their_verkey
        if their_role:
            params["their_role"] = their_role
        if my_seed:
            params["my_seed"] = my_seed
        if my_did:
            params["my_did"] = my_did
        if their_endpoint:
            params["their_endpoint"] = their_endpoint
        if alias:
            params["alias"] = alias
        if their_did:
            params["their_did"] = their_did

        response = await self.admin_POST(f"/connections/create-static",data = params)
        return response


    async def check_connection_ready(self, connection_id, state):
        stored = False
        for connection in self.connections:
            if connection.id == connection_id:
                try:
                    await asyncio.wait_for(connection.detect_state_ready(state), 5)
                    return True
                except asyncio.TimeoutError:
                    response = await self.get_connection(connection_id)
                    return state == response["state"]
        if not stored:
            try:
                response = await self.get_connection(connection_id)
                connection = Connection(response["connection_id"], response["state"])
                self.connections.append(connection)
                await asyncio.wait_for(connection.detect_state_ready(state), 5)
                return True
            except asyncio.TimeoutError:
                return False


    async def is_active(self, connection_id):
        is_active = await self.check_connection_ready(connection_id, 'active')
        if not is_active:
            logger.error(f"Connection {connection_id} not active")
            raise Exception("Connection must be active to send a credential")
        return



from .base import BaseController
from aiohttp import ClientSession
import logging

logger = logging.getLogger("aries_controller.ledger")

class LedgerController(BaseController):

    def __init__(self, admin_url: str, client_session: ClientSession):
        super().__init__(admin_url, client_session)
        self.base_url = "/ledger"

    async def register_nym(self, did, verkey, role: str = None, alias: str = None):
        params = {
            "did": did,
            "verkey": verkey
        }
        if role:
            params["role"] = role
        if alias:
            params["alias"] = alias

        return await self.admin_POST(f"{self.base_url}/register-nym", params=params)


    async def get_did_verkey(self, did):
        params = {
            "did": did
        }
        return await self.admin_GET(f"{self.base_url}/did-verkey", params=params)

    async def get_did_endpoint(self, did):
        params = {
            "did": did
        }
        return await self.admin_GET(f"{self.base_url}/did-endpoint", params=params)

    async def get_taa(self):
        return await self.admin_GET(f"{self.base_url}/taa")

    async def accept_taa(self, data):
        return await self.admin_POST(f"{self.base_url}/taa/accept", data=data)





from .base import BaseController
from aiohttp import ClientSession
import logging

logger = logging.getLogger("aries_controller.credentials")

class CredentialController(BaseController):

    def __init__(self, admin_url: str, client_session: ClientSession):
        super().__init__(admin_url, client_session)

    async def get_by_id(self, credential_id):
        return await self.admin_GET(f"/credential/{credential_id}")

    async def get_credential_mime_types(self, credential_id):
        return await self.admin_GET(f"/credential/mime-types/{credential_id}")

    async def remove_credential(self, credential_id):
        return await self.admin_POST(f"/credential/{credential_id}/remove")

    async def get_all(self, wql_query: str = None, count: int = None, start: int = None):
        params = {}
        if wql_query:
            params["wql"] = wql_query
        if count:
            params["count"] = count
        if start:
            params["start"] = start

        return await self.admin_GET("/credentials", params=params)


from .base import BaseController
from aiohttp import ClientSession
import logging

logger = logging.getLogger("aries_controller.schema")


class SchemaController(BaseController):

    def __init__(self, admin_url: str, client_session: ClientSession):
        super().__init__(admin_url, client_session)
        self.base_url = "/schemas"

    async def get_by_id(self, schema_id):

        response = await self.admin_GET(f"{self.base_url}/{schema_id}")
        return response

    async def get_created_schema(self, schema_id=None, schema_issuer_did=None, schema_name=None, schema_version=None):

        params = {}
        if schema_id:
            params["schema_id"] = schema_id
        if schema_issuer_did:
            params["schema_issuer_did"] = schema_issuer_did
        if schema_name:
            params["schema_name"] = schema_name
        if schema_version:
            params["schema_version"] = schema_version

        response = await self.admin_GET(f"{self.base_url}/created", params=params)

        return response

    async def write_schema(self, schema_name, attributes, schema_version):

        schema_body = {
            "schema_name": schema_name,
            "attributes": attributes,
            "schema_version": schema_version
        }

        response = await self.admin_POST(f"{self.base_url}", schema_body)
        return response









import asyncio

class Connection:

    def __init__(self, id, state):
        self.id = id
        self.state = state
        self.future_state = None
        self.future = asyncio.Future()

    async def detect_state_ready(self, future_state):
        if self.state == future_state:
            return True
        else:
            self.future_state = future_state
            await self.future

    def update_state(self, state):
        self.state = state
        if state == self.future_state:
            self.future.set_result(True)

    @property
    def connection_ready(self):
        return self.state == "active"


from .base import BaseController
from aiohttp import ClientSession
import logging

logger = logging.getLogger("aries_controller.proof")


class ServerController(BaseController):

    def __init__(self, admin_url: str, client_session: ClientSession):
        super().__init__(admin_url, client_session)

    async def get_plugins(self):
        return await self.admin_GET('/plugins')

    async def get_status(self):
        return await self.admin_GET('/status')

    async def reset_status(self):
        return await self.admin_POST('/status/reset')

    async def get_features(self, query: str = None):
        params = {}
        if query:
            params["query"] = query

        return await self.admin_GET('/features', params=params)



import functools
import json
import os
import sys
from timeit import default_timer

import prompt_toolkit
from prompt_toolkit.application import run_in_terminal
#from prompt_toolkit.eventloop.defaults import use_asyncio_event_loop
from prompt_toolkit.patch_stdout import patch_stdout
from prompt_toolkit.shortcuts import ProgressBar

import pygments
from pygments.filter import Filter
from pygments.lexer import Lexer
from pygments.lexers.data import JsonLdLexer
from prompt_toolkit.formatted_text import FormattedText, PygmentsTokens


COLORIZE = bool(os.getenv("COLORIZE", True))


class PrefixFilter(Filter):
    def __init__(self, **options):
        Filter.__init__(self, **options)
        self.prefix = options.get("prefix")

    def lines(self, stream):
        line = []
        for ttype, value in stream:
            if "\n" in value:
                parts = value.split("\n")
                value = parts.pop()
                for part in parts:
                    line.append((ttype, part))
                    line.append((ttype, "\n"))
                    yield line
                    line = []
            line.append((ttype, value))
        if line:
            yield line

    def filter(self, lexer, stream):
        if isinstance(self.prefix, str):
            prefix = ((pygments.token.Generic, self.prefix),)
        elif self.prefix:
            prefix = self.prefix
        else:
            prefix = ()
        for line in self.lines(stream):
            yield from prefix
            yield from line


def print_lexer(
    body: str, lexer: Lexer, label: str = None, prefix: str = None, indent: int = None
):
    if COLORIZE:
        prefix_str = prefix + " " if prefix else ""
        if prefix_str or indent:
            prefix_body = prefix_str + " " * (indent or 0)
            lexer.add_filter(PrefixFilter(prefix=prefix_body))
        tokens = list(pygments.lex(body, lexer=lexer))
        if label:
            fmt_label = [("fg:ansimagenta", label)]
            if prefix_str:
                fmt_label.insert(0, ("", prefix_str))
            print_formatted(FormattedText(fmt_label))
        print_formatted(PygmentsTokens(tokens))
    else:
        print_ext(body, label=label, prefix=prefix)


def print_json(data, label: str = None, prefix: str = None, indent: int = 2):
    if isinstance(data, str):
        data = json.loads(data)
    data = json.dumps(data, indent=2)
    prefix_str = prefix or ""
    print_lexer(data, JsonLdLexer(), label=label, prefix=prefix_str, indent=indent)


def print_formatted(*args, **kwargs):
    prompt_toolkit.print_formatted_text(*args, **kwargs)


def print_ext(
    *msg,
    color: str = None,
    label: str = None,
    prefix: str = None,
    indent: int = None,
    **kwargs,
):
    prefix_str = prefix or ""
    if indent:
        prefix_str += " " * indent
    if color and COLORIZE:
        msg = [(color, " ".join(map(str, msg)))]
        if prefix_str:
            msg.insert(0, ("", prefix_str + " "))
        if label:
            msg.insert(0, ("fg:ansimagenta", label + "\n"))
        print_formatted(FormattedText(msg), **kwargs)
        return
    if label:
        print(label, **kwargs)
    if prefix_str:
        msg = (prefix_str, *msg)
    print(*msg, **kwargs)


def output_reader(handle, callback, *args, **kwargs):
    for line in iter(handle.readline, b""):
        if not line:
            break
        run_in_terminal(functools.partial(callback, line, *args))


def log_msg(*msg, color="fg:ansimagenta", **kwargs):
    run_in_terminal(lambda: print_ext(*msg, color=color, **kwargs))


def log_json(data, **kwargs):
    run_in_terminal(lambda: print_json(data, **kwargs))


def log_status(status: str, **kwargs):
    log_msg(f"\n{status}", color="bold", **kwargs)


def flatten(args):
    for arg in args:
        if isinstance(arg, (list, tuple)):
            yield from flatten(arg)
        else:
            yield arg


def prompt_init():
    if hasattr(prompt_init, "_called"):
        return
    prompt_init._called = True
    use_asyncio_event_loop()


async def prompt(*args, **kwargs):
    prompt_init()
    with patch_stdout():
        try:
            while True:
                tmp = await prompt_toolkit.prompt(*args, async_=True, **kwargs)
                if tmp:
                    break
            return tmp
        except EOFError:
            return None


async def prompt_loop(*args, **kwargs):
    while True:
        option = await prompt(*args, **kwargs)
        yield option


class DurationTimer:
    def __init__(self, label: str = None, callback=None):
        self.callback = callback
        self.duration = None
        self.label = label
        self.last_error = None
        self.total = 0.0
        self.init_time = self.now()
        self.start_time = None
        self.stop_time = None
        self.running = False

    @classmethod
    def now(cls):
        return default_timer()

    def start(self):
        self.start_time = self.now()
        self.running = True

    def stop(self):
        if not self.running:
            return
        self.stop_time = self.now()
        self.duration = self.stop_time - self.start_time
        self.running = False
        self.total += self.duration
        if self.callback:
            self.callback(self)

    def cancel(self):
        self.running = False

    def reset(self):
        self.duration = None
        self.total = 0.0
        self.last_error = None
        restart = False
        if self.running:
            self.stop()
            restart = True
        self.start_time = None
        self.stop_time = None
        if restart:
            self.start()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, err_type, err_value, err_tb):
        self.last_error = err_value
        self.stop()


def log_timer(label: str, show: bool = True, logger=None, **kwargs):
    logger = logger or log_msg
    cb = (
        (
            lambda timer: timer.last_error
            or logger(timer.label, f"{timer.duration:.2f}s", **kwargs)
        )
        if show
        else None
    )
    return DurationTimer(label, cb)


def progress(*args, **kwargs):
    return ProgressBar(*args, **kwargs)


def require_indy():
    try:
        from indy.libindy import _cdll

        _cdll()
    except ImportError:
        print("python3-indy module not installed")
        sys.exit(1)
    except OSError:
        print("libindy shared library could not be loaded")
        sys.exit(1)

# Used to extract a DID from a schema or credential definition id
def extract_did(id):
    split = id.split(":")
    if len(split) > 3:
        return split[0]
    else:
        raise Exception(f"ID {id} is not is the correct format")


def get_schema_details(schema_id):
    details = schema_id.split(":")
    if len(details) == 4:
        return {
            "schema_id": schema_id,
            "schema_name": details[2],
            "schema_version": details[3],
            "schema_issuer_did": details[0]
        }
    else:
        raise Exception(f"ID {id} is not is the correct")


from .base import BaseController
from aiohttp import ClientSession
import logging

logger = logging.getLogger("aries_controller.messaging")

class MessagingController(BaseController):

    def __init__(self, admin_url: str, client_session: ClientSession):
        super().__init__(admin_url, client_session)


    def default_handler(self, payload):
        logger.debug("Message Received ", payload)

    async def send_message(self, connection_id, msg):
        response = await self.admin_POST(f"/connections/{connection_id}/send-message", {
            "content": msg,
        })
        return response

    async def trust_ping(self, connection_id: str, msg: str):
        response = await self.admin_POST(f"/connections/{connection_id}/send-ping", {"content": msg})
        return response



from aiohttp import (
    web,
    ClientSession,
    ClientRequest,
)
from pubsub import pub

from .controllers.connections import ConnectionsController
from .controllers.messaging import MessagingController
from .controllers.schema import SchemaController
from .controllers.wallet import WalletController
from .controllers.definitions import DefinitionsController
from .controllers.issuer import IssuerController
from .controllers.proof import ProofController
from .controllers.ledger import LedgerController
from .controllers.credential import CredentialController
from .controllers.server import ServerController

import logging

logger = logging.getLogger("aries_controller")

class AriesAgentController:

    ## TODO rethink how to initialise. Too many args?
    ## is it important to let users config connections/issuer etc
    def __init__(self, webhook_host: str, webhook_port: int, admin_url: str, webhook_base: str = "",
                 connections: bool = True, messaging: bool = True, issuer: bool = True, api_key: str = None):

        self.webhook_site = None
        self.admin_url = admin_url
        if webhook_base:
            self.webhook_base = webhook_base
        else:
            self.webhook_base = ""
        self.webhook_host = webhook_host
        self.webhook_port = webhook_port
        self.connections_controller = None

        if api_key:
            headers = {"X-API-Key": api_key}
            self.client_session: ClientSession = ClientSession(headers=headers)
        else:
            self.client_session: ClientSession = ClientSession()

        if connections:
            self.connections = ConnectionsController(self.admin_url, self.client_session)
        if messaging:
            self.messaging = MessagingController(self.admin_url, self.client_session)

        self.proofs = ProofController(self.admin_url, self.client_session)
        self.ledger = LedgerController(self.admin_url, self.client_session)
        self.credentials = CredentialController(self.admin_url, self.client_session)
        self.server = ServerController(self.admin_url, self.client_session)
        if issuer:
            self.schema = SchemaController(self.admin_url, self.client_session)
            self.wallet = WalletController(self.admin_url, self.client_session)
            self.definitions = DefinitionsController(self.admin_url, self.client_session)
            self.issuer = IssuerController(self.admin_url, self.client_session, self.connections,
                                           self.wallet, self.definitions)



    def register_listeners(self, listeners, defaults=True):
        if defaults:
            if self.connections:
                pub.subscribe(self.connections.default_handler, "connections")
            if self.messaging:
                pub.subscribe(self.messaging.default_handler, "basicmessages")
            if self.proofs:
                pub.subscribe(self.proofs.default_handler, "present_proof")


        for listener in listeners:
            pub.subscribe(listener["handler"], listener["topic"])

    async def listen_webhooks(self):
        app = web.Application()
        app.add_routes([web.post(self.webhook_base + "/topic/{topic}/", self._receive_webhook)])
        runner = web.AppRunner(app)
        await runner.setup()
        self.webhook_site = web.TCPSite(runner, self.webhook_host, self.webhook_port)
        await self.webhook_site.start()

    async def _receive_webhook(self, request: ClientRequest):
        topic = request.match_info["topic"]
        payload = await request.json()
        await self.handle_webhook(topic, payload)
        return web.Response(status=200)

    async def handle_webhook(self, topic, payload):
        logging.debug(f"Handle Webhook - {topic}", payload)
        pub.sendMessage(topic, payload=payload)
        return web.Response(status=200)


    async def terminate(self):
        await self.client_session.close()
        if self.webhook_site:
            await self.webhook_site.stop()






from aiohttp import (
    web,
    ClientSession,
    ClientRequest,
    ClientResponse,
    ClientError,
    ClientTimeout,
)

import asyncio
import logging

import json

from ..helpers.utils import log_msg

EVENT_LOGGER = logging.getLogger("event")

class repr_json:
    def __init__(self, val):
        self.val = val

    def __repr__(self) -> str:
        if isinstance(self.val, str):
            return self.val
        return json.dumps(self.val, indent=4)


class BaseController:

    def __init__(self, admin_url: str, client_session):
        self.admin_url = admin_url
        self.client_session = client_session
        self.color = None
        self.prefix = None

    def log(self, *msg, **kwargs):
        self.handle_output(*msg, **kwargs)

    @property
    def prefix_str(self):
        if self.prefix:
            return f"{self.prefix:10s} |"

    def handle_output(self, *output, source: str = None, **kwargs):
        end = "" if source else "\n"
        if source == "stderr":
            color = "fg:ansired"
        elif not source:
            color = self.color or "fg:ansiblue"
        else:
            color = None
        log_msg(*output, color=color, prefix=self.prefix_str, end=end, **kwargs)


    async def admin_request(
        self, method, path, data=None, text=False, params=None
    ) -> ClientResponse:
        params = {k: v for (k, v) in (params or {}).items() if v is not None}
        async with self.client_session.request(
            method, self.admin_url + path, json=data, params=params
        ) as resp:
            resp.raise_for_status()
            resp_text = await resp.text()
            if not resp_text and not text:
                return None
            if not text:
                try:
                    return json.loads(resp_text)
                except json.JSONDecodeError as e:
                    raise Exception(f"Error decoding JSON: {resp_text}") from e
            return resp_text


    async def admin_GET(self, path, text=False, params=None) -> ClientResponse:
        try:
            EVENT_LOGGER.debug("Controller GET %s request to Agent", path)
            response = await self.admin_request("GET", path, None, text, params)
            EVENT_LOGGER.debug(
                "Response from GET %s received: \n%s", path, repr_json(response),
            )
            return response
        except ClientError as e:
            self.log(f"Error during GET {path}: {str(e)}")
            raise

    async def admin_POST(
            self, path, data=None, text=False, params=None
    ) -> ClientResponse:
        try:
            EVENT_LOGGER.debug(
                "Controller POST %s request to Agent%s",
                path,
                (" with data: \n{}".format(repr_json(data)) if data else ""),
            )
            response = await self.admin_request("POST", path, data, text, params)
            EVENT_LOGGER.debug(
                "Response from POST %s received: \n%s", path, repr_json(response),
            )
            return response
        except ClientError as e:
            self.log(f"Error during POST {path}: {str(e)}")
            raise





from .base import BaseController
from aiohttp import ClientSession
import logging

logger = logging.getLogger("aries_controller.wallet")

class WalletController(BaseController):

    def __init__(self, admin_url: str, client_session: ClientSession):
        super().__init__(admin_url, client_session)
        self.base_url = "/wallet"

    async def get_dids(self):
        return await self.admin_GET(f"{self.base_url}/did")


    async def create_did(self):
        return await self.admin_POST(f"{self.base_url}/did/create")

    async def get_public_did(self):
        return await self.admin_GET(f"{self.base_url}/did/public")

    async def assign_public_did(self, did):
        params = {
            "did": did
        }
        return await self.admin_POST(f"{self.base_url}/did/public", params=params)

    async def get_tag_policy(self, credential_definition_id):
        return await self.admin_GET(f"{self.base_url}/tag-policy/{credential_definition_id}")

    async def set_tag_policy(self, credential_definition_id):
        return await self.admin_POST(f"{self.base_url}/tag-policy/{credential_definition_id}")